#!/usr/bin/env python
"""
Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import boto3
import sys
import time
import argparse
import re
import json
import random
import string

from collections import OrderedDict
from botocore.exceptions import ClientError
from six.moves import input as raw_input
def assume_role(aws_account_number, role_name):
    """
    Assumes the provided role in each account and returns a SecurityHub client
    :param aws_account_number: AWS Account Number
    :param role_name: Role to assume in target account
    :param aws_region: AWS Region for the Client call, not required for IAM calls
    :return: SecurityHub client in the specified AWS Account and Region
    """

    # Beginning the assume role process for account
    sts_client = boto3.client('sts')
    
    # Get the current partition
    partition = sts_client.get_caller_identity()['Arn'].split(":")[1]
    
    response = sts_client.assume_role(
        RoleArn='arn:{}:iam::{}:role/{}'.format(
            partition,
            aws_account_number,
            role_name
        ),
        RoleSessionName='EnableSecurityHub'
    )
    
    # Storing STS credentials
    session = boto3.Session(
        aws_access_key_id=response['Credentials']['AccessKeyId'],
        aws_secret_access_key=response['Credentials']['SecretAccessKey'],
        aws_session_token=response['Credentials']['SessionToken']
    )

    print("Assumed session for {}.".format(
        aws_account_number
    ))

    return session

def get_parent_members(sh_client, aws_region):
    """
    Returns a list of current members of the SecurityHub parent account
    :param aws_region: AWS Region of the SecurityHub parent account
    :param detector_id: DetectorId of the SecurityHub parent account in the AWS Region
    :return: dict of AwsAccountId:RelationshipStatus
    """

    member_dict = dict()

    results = sh_client.list_members(
        OnlyAssociated=False
    )
    
    for member in results['Members']:
        member_dict.update({member['AccountId']: member['MemberStatus']})
        
    while results.get("NextToken"):
        results = sh_client.list_members(
            OnlyAssociated=False,
            NextToken=results['NextToken']
        )
        
        for member in results['Members']:
            member_dict.update({member['AccountId']: member['MemberStatus']})
            
    return member_dict

if __name__ == '__main__':
    
    # Setup command line arguments
    parser = argparse.ArgumentParser(description='Link AWS Accounts to central SecurityHub Account')
    parser.add_argument('--parent_account', type=str, required=True, help="AccountId for Central AWS Account")
    parser.add_argument('input_file', type=argparse.FileType('r'), help='Path to CSV file containing the list of account IDs and Email addresses')
    parser.add_argument('--assume_role', type=str, required=True, help="Role Name to assume in each account")
    parser.add_argument('--enabled_regions', type=str, help="comma separated list of regions to enable SecurityHub. If not specified, all available regions enabled")
    parser.add_argument('--enable_standards', type=str, required=False,help="comma separated list of standards ARNs to enable ( i.e. arn:aws:securityhub:::ruleset/cis-aws-foundations-benchmark/v/1.2.0 )")
    args = parser.parse_args()

    # Validate parent accountId
    if not re.match(r'[0-9]{12}',args.parent_account):
        raise ValueError("Parent AccountId is not valid")
    
    # Generate dict with account & email information
    aws_account_dict = OrderedDict()
    
    # Notify on Config dependency if standards are enabled
    if args.enable_standards:
        print(
        '''
        *****************************************************************************************************************************************************************************************
        *      By turning on this Standards you will enable security evaluations to run. For current pricing and example scenarios please refer to the current AWS Security Hub pricing.        *
        *      Important: You must enable AWS Config for all resources in each AWS Region where you will be running a Standard. If Config is not already enabled it will be enabled and         *
        *      configured in each region.                                                                                                                                                       *
        *                                                                                                                                                                                       *
        *      In addition to AWS Security Hub charges, you will also incur charges for the Configuration Items recorded by AWS Config, as per the AWS Config pricing. These charges are        *
        *      separate from (and not included in) AWS Security Hub pricing.                                                                                                                    *
        *****************************************************************************************************************************************************************************************
        
        Continue?(yes/no):
        '''
        )
        notify_config_response = ''
        if 'yes' not in raw_input(notify_config_response).lower():
            print("Exiting..")
            raise SystemExit(0)
     
    for acct in args.input_file.readlines():
        split_line = acct.rstrip().split(",")
        if len(split_line) < 2:
            print("Unable to process line: {}".format(acct))
            continue
            
        if not re.match(r'[0-9]{12}', str(split_line[0])):
            print("Invalid account number {}, skipping".format(split_line[0]))
            continue
            
        aws_account_dict[split_line[0]] = split_line[1]

    # Check length of accounts to be processed
    if len(aws_account_dict.keys()) > 1000:
        raise Exception("Only 1000 accounts can be linked to a single parent account")
    
    # Getting SecurityHub regions
    session = boto3.session.Session()

    securityhub_regions = []
    if args.enabled_regions:
        securityhub_regions = [str(item) for item in args.enabled_regions.split(',')]
        print("Enabling members in these regions: {}".format(securityhub_regions))
    else:
        securityhub_regions = session.get_available_regions('securityhub')
        print("Enabling members in all available SecurityHub regions {}".format(securityhub_regions))
    
    # Check if enable Standards 
    standards_arns = []
    if args.enable_standards:
        standards_arns = [str(item) for item in args.enable_standards.split(',')]
        print("Enabling the following Security Hub Standards for enabled account(s) and region(s): {}".format(standards_arns))
        

    # Processing Parent account
    parent_session = assume_role(args.parent_account, args.assume_role)
    #parent_session = boto3.Session()
    parent_clients = {}
    members = {}
    for aws_region in securityhub_regions:
        parent_clients[aws_region] = parent_session.client('securityhub', region_name=aws_region)
        try: 
            parent_clients[aws_region].enable_security_hub()
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceConflictException':
                pass
            else:
                print("Error: Unable to enable Security Hub on Parent account in region {}").format(aws_region)
                raise SystemExit(0)

        members[aws_region] = get_parent_members(parent_clients[aws_region], aws_region)

    # Processing accounts to be linked
    failed_accounts = []
    for account in aws_account_dict.keys():
        try:
            session = assume_role(account, args.assume_role)
            
            for aws_region in securityhub_regions:
                print('Beginning {account} in {region}'.format(
                    account=account,
                    region=aws_region
                ))
                
                sh_client = session.client('securityhub', region_name=aws_region)
            
                try:
                    sh_client.enable_security_hub()
                except ClientError as e:
                    if e.response['Error']['Code'] == 'ResourceConflictException':
                        pass
                for standard in standards_arns:
                    sh_client.batch_enable_standards(StandardsSubscriptionRequests=[{'StandardsArn' : standard}])
                    start_time = int(time.time())
                    status = ''
                    while status != 'READY':
                        if (int(time.time()) - start_time) > 100:
                            print("Timeout waiting for READY state enabling standard {standard} in region {region} for account {account}, last state: {status}".format(standard=standard,region=aws_region, account=account, status=status))
                            break
                        enabled_standards = sh_client.get_enabled_standards()
                        for enabled_stanard in enabled_standards['StandardsSubscriptions']:
                            if enabled_stanard['StandardsArn'] == standard:
                                status = enabled_stanard['StandardsStatus']
                    if status == 'READY':
                        print("Finished enabling stanard {} on account {} for region {}".format(standard,account, aws_region))
                                
                    

                if account not in members[aws_region]:
        
                    parent_clients[aws_region].create_members(
                        AccountDetails=[{
                            "AccountId": account,
                            "Email": aws_account_dict[account]
                        }]
                    )
                
                    print('Added Account {monitored} to member list in SecurityHub parent account {parent} for region {region}'.format(
                        monitored=account,
                        parent=args.parent_account,
                        region=aws_region
                    ))
                                
                    start_time = int(time.time())
                    while account not in members[aws_region]:
                        if (int(time.time()) - start_time) > 300:
                            print("Membership did not show up for account {}, skipping".format(account))
                            failed_accounts.append({
                                account: "Membership did not show up for account {} in {}".format(
                                    account,
                                    aws_region
                                )
                            })
                            break
                        
                        time.sleep(5)
                        members[aws_region] = get_parent_members(parent_clients[aws_region], aws_region)

                else:
                    print('Account {monitored} is already a member of {parent} in region {region}'.format(
                        monitored=account,
                        parent=args.parent_account,
                        region=aws_region
                    ))
                print('list of memebers - {members}'.format(members=members))
                if members[aws_region][account] == 'Enabled':
                    # Member is enabled and already being monitored
                    print('Account {account} is already enabled'.format(account=account))
                
                else:
                    start_time = int(time.time())
                    while members[aws_region][account] != 'Enabled':
                        if (int(time.time()) - start_time) > 300:
                            print("Invitation did not show up for account {}, skipping".format(account))
                            failed_accounts.append({
                                account: "Membership did not show up for account {} in {}".format(
                                    account,
                                    aws_region
                                )
                            })
                            break
                        
                        if members[aws_region][account] == 'Created':
                            # Member has been created in the SecurityHub parent account but not invited yet
                            parent_clients[aws_region].invite_members(
                                AccountIds=[account]
                            )
                        
                            print('Invited Account {monitored} to SecurityHub parent account {parent} in region {region}'.format(
                                monitored=account,
                                parent=args.parent_account,
                                region=aws_region
                            ))
        
                        if members[aws_region][account] == 'Invited' or members[aws_region][account] == 'Resigned':
                            # member has been invited so accept the invite
                            
                            response = sh_client.list_invitations()
                            
                            invitation_dict = dict()
                            
                            invitation_id = None
                            for invitation in response['Invitations']:
                                invitation_id = invitation['InvitationId']
                            
                            if invitation_id is not None:
                                sh_client.accept_invitation(
                                    InvitationId=invitation_id,
                                    MasterId=str(args.parent_account)
                                )
                                print('Accepting Account {monitored} to SecurityHub parent account {parent} in region {region}'.format(
                                    monitored=account,
                                    parent=args.parent_account,
                                    region=aws_region
                                ))
    
                        # Refresh the member dictionary
                        members[aws_region] = get_parent_members(parent_clients[aws_region], aws_region)
                        
                    print('Finished {account} in {region}'.format(account=account, region=aws_region))
                    
        except ClientError as e:
            print("Error Processing Account {}".format(account))
            failed_accounts.append({
                account: repr(e)
            })

    if len(failed_accounts) > 0:
        print("---------------------------------------------------------------")
        print("Failed Accounts")
        print("---------------------------------------------------------------")
        print('failed accounts - {fail}'.format(fail=failed_accounts))
        for account in failed_accounts:
            print("{}: \n\t{}".format(
                account.keys(),
                account[account.keys()]
            ))
            print("---------------------------------------------------------------")
