#!/usr/bin/env python
"""
Copyright 2018 Amazon.com, Inc. or its affiliates.
All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License").
You may not use this file except in compliance with the License.
A copy of the License is located at

   http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file.
This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.

This script orchestrates the enablement and centralization of vpc_flow_logs2 across an enterprise of AWS accounts.
It takes in a list of AWS Account Numbers, iterates through each account and region to enable vpc_flow_logs2.
It creates each account as a Member in the vpc_flow_logs2 Parent account.
It invites and accepts the invite for each Member account.
"""

import boto3
import sys
import time
import argparse
import re

from collections import OrderedDict
from botocore.exceptions import ClientError

def assume_role(aws_account_number, role_name):
    """
    Assumes the provided role in each account and returns a vpc_flow_logs2 client
    :param aws_account_number: AWS Account Number
    :param role_name: 0Role to assume in target account
    :param aws_region: AWS Region for the Client call, not required for IAM calls
    :return: vpc_flow_logs client in the specified AWS Account and Region
    """

    # Beginning the assume role process for account
    sts_client = boto3.client('sts')

    # Get the current partition
    partition = sts_client.get_caller_identity()['Arn'].split(":")[1]

    response = sts_client.assume_role(
        RoleArn='arn:{}:iam::{}:role/{}'.format(
            partition,
            aws_account_number,
            role_name
        ),
        RoleSessionName='Enablevpc_flow_logs'
    )

    # Storing STS credentials
    session = boto3.Session(
        aws_access_key_id=response['Credentials']['AccessKeyId'],
        aws_secret_access_key=response['Credentials']['SecretAccessKey'],
        aws_session_token=response['Credentials']['SessionToken']
    )

    print("Assumed session for {}.".format(
        aws_account_number
    ))

    return session


if __name__ == '__main__':

    # Setup command line arguments
    parser = argparse.ArgumentParser(description='Link AWS Accounts to central vpc_flow_logs Account')
    parser.add_argument('--parent_account', type=str, required=True, help="AccountId for Central AWS Account")
    parser.add_argument('input_file', type=argparse.FileType('r'), help='Path to CSV file containing the list of account IDs and S3 resources')
    parser.add_argument('--assume_role', type=str, required=True, help="Role Name to assume in each account")
    parser.add_argument('--enabled_regions', type=str, help="comma separated list of regions to enable vpc_flow_logs. If not specified, all available regions enabled")
    parser.add_argument('--centrals3bucket', type=str, required=True, help="S3 bucket name where all vpc flows logs will be stored across all the accounts")    
    args = parser.parse_args()

    # Validate parent accountId
    if not re.match(r'[0-9]{12}', args.parent_account):
        raise ValueError("Parent AccountId is not valid")
    parent_account = args.parent_account
    central_s3_bucket = args.centrals3bucket
    # Generate dict with account & email information

    aws_account_dict = OrderedDict()

    for acct in args.input_file.readlines():
        split_line = acct.rstrip().split(",")
        if len(split_line) < 2:
            print("Unable to process line: {}".format(acct))
            continue

        if not re.match(r'[0-9]{12}', str(split_line[0])):
            print("Invalid account number {}, skipping".format(split_line[0]))
            continue

        aws_account_dict[split_line[0]] = split_line[1]

    # Getting vpc_flow_logs regions
    session = boto3.session.Session()

    vpc_regions = []
    vpc_regions = [str(item) for item in args.enabled_regions.split(',')]
    print("Enabling VPCFlowLogs in these regions: {}".format(vpc_regions))



    # Associating member accounts
    failed_accounts = []
    for account in aws_account_dict.keys():
        for aws_region in vpc_regions:
            try:
                child_session = assume_role(account, args.assume_role)
                vpc_flow_logs_client = child_session.client('ec2', region_name=aws_region)
                List_VPCs = vpc_flow_logs_client.describe_vpcs()['Vpcs']
                vpcid_list = []
                for vpcs in List_VPCs:
                    vpcid = vpcs['VpcId']
                    vpcid_list.append(vpcid)
                print('VPCs in the region {region} for the account {account}, are - {vpcs}'.format(region=aws_region,account=account,vpcs=vpcid_list))

                List_Flow_logs = vpc_flow_logs_client.describe_flow_logs()['FlowLogs']
                flowlog_vpcid_list = []
                for flowlogs in List_Flow_logs:
                    vpcid_flowlogs = flowlogs['ResourceId']
                    flowlog_vpcid_list.append(vpcid_flowlogs)
                
                flowlogs_vpcid_list_to_enable = []

                for new_vpcIds in vpcid_list:
                    if new_vpcIds not in flowlog_vpcid_list:
                        flowlogs_vpcid_list_to_enable.append(new_vpcIds)

                print('==========VPC ID LIST ON WHICH NEED TO ENABLE THE VPC_FLOW_LOGS===========')
                print(flowlogs_vpcid_list_to_enable)
                print('VPCs in the region: {region} for the account: {account}, for which flows logs are not enabled - {vpcs}'.format(region=aws_region,account=account,vpcs=flowlogs_vpcid_list_to_enable))

                response = vpc_flow_logs_client.create_flow_logs(
                    ResourceType='VPC',
                    TrafficType='ALL',
                    LogDestinationType='s3',
                    LogDestination='arn:aws:s3:::{s3bucket}/'.format(s3bucket=central_s3_bucket),
                    ResourceIds=flowlogs_vpcid_list_to_enable
                    )
                print('Successfully enable the vpcflow logs for vpcs: {vpcs} ,in the region: {region}'.format(vpcs=flowlogs_vpcid_list_to_enable,region=aws_region))

            except ClientError as err:
                errormessage=err.response
                print(errormessage['Error']['Message'])
                if err.response['ResponseMetadata']['HTTPStatusCode'] == 403:
                    print("Failed to enable vpc_flow_logs in Parent account for region: {} due to an authentication error.  Either your credentials are not correctly configured or the region is an OptIn region that is not enabled on the parent account.  Skipping {} and attempting to continue".format(aws_region,aws_region))
                    failed_accounts.append(account)

    print("Enabling VPC Flow Logs in the Security Parent Account Now")
    for aws_region in vpc_regions:
        try:
                child_session = assume_role(parent_account, args.assume_role)
                vpc_flow_logs_client = child_session.client('ec2', region_name=aws_region)
                List_VPCs = vpc_flow_logs_client.describe_vpcs()['Vpcs']
                vpcid_list = []
                for vpcs in List_VPCs:
                    vpcid = vpcs['VpcId']
                    vpcid_list.append(vpcid)
                print('VPCs in the region: {region} for the account: {account}, are - {vpcs}'.format(region=aws_region,account=account,vpcs=vpcid_list))

                List_Flow_logs = vpc_flow_logs_client.describe_flow_logs()['FlowLogs']
                flowlog_vpcid_list = []
                for flowlogs in List_Flow_logs:
                    vpcid_flowlogs = flowlogs['ResourceId']
                    flowlog_vpcid_list.append(vpcid_flowlogs)
                
                flowlogs_vpcid_list_to_enable = []

                for new_vpcIds in vpcid_list:
                    if new_vpcIds not in flowlog_vpcid_list:
                        flowlogs_vpcid_list_to_enable.append(new_vpcIds)

                print('==========VPC ID LIST ON WHICH NEED TO ENABLE THE VPC_FLOW_LOGS===========')
                print(flowlogs_vpcid_list_to_enable)
                print('VPCs in the region: {region} for the account: {account}, for which flows logs are not enabled - {vpcs}'.format(region=aws_region,account=account,vpcs=flowlogs_vpcid_list_to_enable))

                response = vpc_flow_logs_client.create_flow_logs(
                    ResourceType='VPC',
                    TrafficType='ALL',
                    LogDestinationType='s3',
                    LogDestination='arn:aws:s3:::{s3bucket}/'.format(s3bucket=central_s3_bucket),
                    ResourceIds=flowlogs_vpcid_list_to_enable
                    )
                print('Successfully enable the vpcflow logs for vpcs: {} ,in the region: {}'.format(flowlogs_vpcid_list_to_enable,aws_region))

        except ClientError as err:
                errormessage=err.response
                print(errormessage['Error']['Message'])
                if err.response['ResponseMetadata']['HTTPStatusCode'] == 403:
                    print("Failed to enable vpc_flow_logs in Parent account for region: {} due to an authentication error.  Either your credentials are not correctly configured or the region is an OptIn region that is not enabled on the parent account.  Skipping {} and attempting to continue".format(aws_region,aws_region))
                    failed_accounts.append(parent_account)


    if len(failed_accounts) > 0:
        print("---------------------------------------------------------------")
        print("Failed Accounts")
        print("---------------------------------------------------------------")
        for account in failed_accounts:
            print("{}".format(account))
            print("---------------------------------------------------------------")
